<?php

namespace Snaju\Inception\Http\Request\Validators;

use Snaju\Inception\Exception\InvalidArgumentException;
use Snaju\Inception\Http\Request\Validators\Rules\AlphaRule;
use Snaju\Inception\Http\Request\Validators\Rules\BailRule;
use Snaju\Inception\Http\Request\Validators\Rules\MaxLenRule;
use Snaju\Inception\Http\Request\Validators\Rules\MinLenRule;
use Snaju\Inception\Http\Request\Validators\Rules\NumericRule;
use Snaju\Inception\Http\Request\Validators\Rules\RequiredRule;
use Snaju\Inception\Http\Request\WebRequest;
use Snaju\Inception\InceptionCore;
use Snaju\Inception\Util\Loaders\Collection;
use Snaju\Inception\Http\Request\Validators\ValidatorInterface;

class Validator implements ValidatorInterface
{

    private WebRequest $request;

    private array $rules = [];

    private array $data = [];

    private Collection $errors;

    private Collection $defaultRules;

    private bool $usesRequestData;

    private function addError(string $key, string $error)
    {
        $this->errors->add($key, $error);
    }

    private function sendErrorResponse()
    {

    }

    public function __construct($rules = [], $data = null, bool $sendErrorResponses = true)
    {
        $this->errors = new Collection();
        $this->request = InceptionCore::getRequest();
        $this->defaultRules = new Collection([
            'alpha'         => AlphaRule::class,
            'bail'          => BailRule::class,
            'max'           => MaxLenRule::class,
            'min'           => MinLenRule::class,
            'numeric'       => NumericRule::class,
            'required'      => RequiredRule::class,
        ]);

        // If data variable is empty, validate all data in request
        // Else, just validate the data passed to the class
        if (empty($data)) {
            $this->data = $this->request->toArray();
            $this->usesRequestData = true;
        }
        else {
            $this->data = $data;
            $this->usesRequestData = false;
        }

        if (count($rules) > 0) {
            $this->validate($this->rules, $this->data, $sendErrorResponses);
        }
    }

    public function validate($rules = [], $data = [], $sendErrorResponses = false)
    {

        // Grab all the rules and create an array of rules and values
        return response()->json([
            'rules' => $rules,
            'data' => $data,
        ]);


        // Grab all the data



        // Validate

        // If an id doesn't have any dot syntax, just look at the data as is.
        // But if using request data, look into the body of the request.

        // If id has dot syntax, just look at the data as is














        if ($sendErrorResponses && $this->hasErrors()) {
            // TODO: in the future take into consideration the Accept to check what type of response to send instead of just JSON responses
            return response()->json(['errors' => $this->getErrors()], 400);
        }
    }

    public function hasErrors(): bool
    {
        $errors = false;

        if ($this->errors->length() > 0) {
            return true;
        }

        return $errors;
    }

    public function getErrors(): array
    {
        return $this->errors->toArray();
    }

    public function sendErrors($errors)
    {
        // TODO: Implement sendErrors() method.
    }

    public function addRule(string $identifier, $rule)
    {
        if ($rule instanceof RuleInterface) {
            $this->defaultRules->add($identifier, $rule::class);
        }
        else {
            throw new InvalidArgumentException('The validation Rule must implement the RuleInterface');
        }
    }

}